/* importListDialog.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "importListDialog.h"
#include "importList.h"
#include "importDialog.h"
#include "importingDialog.h"
#include "widgetSize.h"
#include "format.h"
#include "fix_capitalization.h"
#include "artist.h"
#include "database.h"
#include "dataCell.h"
#include "string_utils.h"
#include "fieldHandle.h"
#include "playlist.h"
#include "playlistDialog.h"

static string
clean_spaces(const string &str) {
  string result;
  bool last_space = true;
  for (int i = 0; i < str.length(); i++) {
    if (str[i] == '-' || isspace(str[i])) {
      if (!last_space) {
	result += ' ';
      }
      last_space = true;
    } else {
      last_space = false;
      result += str[i];
    }
  }

  return result;
}

ImportListDialog::
ImportListDialog(const Dataset &data, ImportDialog *import_dialog,
		 const Volume &volume, Disposition disp,
		 int convert_to_type, const string &root_prefix,
		 ImportList *good_files, ImportList *bad_files) :
  PalomaDialog(data),
  _import_dialog(import_dialog),
  _volume(volume),
  _disp(disp),
  _convert_to_type(convert_to_type),
  _root_prefix(root_prefix),
  _good_files(good_files),
  _bad_files(bad_files),
  _good_label("Matching files\nClick on a column heading to define the column's meaning."),
  _bad_label("Unmatching files\nThese files didn't match the regular expression or had an unknown type."),
  _playlist_check("Popup playlist for imported songs"),
  _ok_button("OK"),
  _cancel_button("Cancel")
{
  assert(_good_files != NULL);
  assert(_bad_files != NULL);

  // We need to be notified if the import_dialog is deleted, so we can
  // clear its pointer.
  if (_import_dialog != NULL) {
    connect_to_method(_import_dialog->deleted, this, &import_dialog_deleted);
  }

  if (_disp == D_leave) {
    // If we'll be leaving the files where they are, we need to
    // extract out of the root_prefix just the part that's relative to
    // the volume definition.
    string volume_def = _volume.get_definition();
    assert(_root_prefix.substr(0, volume_def.length()) == volume_def);

    _volume_prefix = _root_prefix.substr(volume_def.length());
    while (!_volume_prefix.empty() && _volume_prefix[0] == '/') {
      // It shouldn't begin with a slash.
      _volume_prefix = _volume_prefix.substr(1);
    }
  }

  _good_clist = _good_files->make_clist();
  assert(_good_clist != NULL);
  _bad_clist = NULL;
  if (_bad_files->get_num_rows() > 0) {
    _bad_clist = _bad_files->make_clist();
    _bad_clist->column_titles_passive();
  }

  layout_window();
  setup();

  // Now build the popup menu for the columns of the good CList.
  _popup_menu = new Gtk_Menu;

  Gtk_MenuItem *clean_spaces = 
    new Gtk_MenuItem("Clean spaces");
  clean_spaces->show();
  _popup_menu->append(manage(clean_spaces));
  connect_to_method(clean_spaces->activate, this, 
		    &picked_popup_clean_spaces);
  
  Gtk_MenuItem *fix_capitalization = 
    new Gtk_MenuItem("Fix capitalization");
  fix_capitalization->show();
  _popup_menu->append(manage(fix_capitalization));
  connect_to_method(fix_capitalization->activate, this, 
		    &picked_popup_fix_capitalization);
  
  Gtk_MenuItem *sep = new Gtk_MenuItem;
  sep->show();
  _popup_menu->append(manage(sep));
  
  for (int t = 0; t < (int)ImportList::CT_num_types; t++) {
    ImportList::ColumnType type = (ImportList::ColumnType)t;
    Gtk_MenuItem *item = 
      new Gtk_MenuItem(ImportList::get_type_description(type));
    item->show();
    _popup_menu->append(manage(item));
    connect_to_method(item->activate, this, &picked_popup_type, type);
  }

  _popup_colnum = -1;
  connect_to_method(_good_clist->click_column, this, &got_column_click);
}

ImportListDialog::
~ImportListDialog() {
  delete _good_files;
  delete _bad_files;

  // For some reason Gtk complains when I do this.  What's wrong?
  //  delete _popup_menu;

  // Free all the song pointers we might have allocated.
  Songs::iterator si;
  for (si = _songs.begin(); si != _songs.end(); ++si) {
    delete (*si);
  }
}

void ImportListDialog::
got_column_click(gint colnum) {
  _popup_colnum = colnum;
  _popup_menu->popup(NULL, NULL, 1, 0);
}

void ImportListDialog::
picked_popup_clean_spaces() {
  assert(_popup_colnum >= 0 && _popup_colnum < _good_files->_columns.size());
  ImportList::Column &column = _good_files->_columns[_popup_colnum];

  _good_clist->freeze();

  int num_rows = _good_files->get_num_rows();
  for (int r = 0; r < num_rows; r++) {
    ImportList::Row &row = _good_files->_rows[r];
    if (_popup_colnum < row._data.size()) {
      row._data[_popup_colnum] = clean_spaces(row._data[_popup_colnum]);
      _good_clist->set_text(r, _popup_colnum, row._data[_popup_colnum]);
    }
  }

  _good_clist->thaw();
}

void ImportListDialog::
picked_popup_fix_capitalization() {
  assert(_popup_colnum >= 0 && _popup_colnum < _good_files->_columns.size());
  ImportList::Column &column = _good_files->_columns[_popup_colnum];

  _good_clist->freeze();

  int num_rows = _good_files->get_num_rows();
  for (int r = 0; r < num_rows; r++) {
    ImportList::Row &row = _good_files->_rows[r];
    if (_popup_colnum < row._data.size()) {
      row._data[_popup_colnum] = 
	fix_capitalization(row._data[_popup_colnum]);
      _good_clist->set_text(r, _popup_colnum, row._data[_popup_colnum]);
    }
  }

  _good_clist->thaw();
}

void ImportListDialog::
picked_popup_type(ImportList::ColumnType type) {
  assert(_popup_colnum >= 0 && _popup_colnum < _good_files->_columns.size());
  ImportList::Column &column = _good_files->_columns[_popup_colnum];

  column.set_type(type);
  _good_clist->set_column_title(_popup_colnum, column._heading);
}

void ImportListDialog::
ok_button_clicked() {
  collect_songs();
  import_songs();
  if (_playlist_check.get_active()) {
    build_playlist();
  }

  // import_songs() only builds up the records in the SQL database,
  // which is sufficient only if we're leaving the song files where
  // they are.  If we're supposed to be copying or converting them,
  // however, we must do more.
  if (_disp != D_leave) {
    new ImportingDialog(_data, _volume, _convert_to_type, _root_prefix,
			_songs);
  }

  // If we still have an import dialog sitting around, kill it.
  if (_import_dialog != NULL) {
    delete _import_dialog;
  }

  delete_self();
}

void ImportListDialog::
cancel_button_clicked() {
  delete_self();
}

void ImportListDialog::
import_dialog_deleted() {
  _import_dialog = NULL;
}

void ImportListDialog::
layout_window() {
  set_title("Import file list");
  set_default_size(600, 0);

  // The paned window.

  // We only need the paned window if we have a bad_clist.
  if (_bad_clist != NULL) {
    vbox_.pack_start(_paned);
    _paned.show();
  }

  // The "good" files list.
  if (_bad_clist != NULL) {
    _paned.add1(_good_box);
  } else {
    vbox_.pack_start(_good_box);
  }
  _good_box.show();
  _good_box.pack_start(_good_label, false, false, 8);
  _good_label.show();
  request_initial_size(_good_window, 0, 200);
  _good_box.pack_start(_good_window);
  _good_window.show();
  _good_window.add(manage(_good_clist));
  _good_clist->show();

  // The "bad" files list.
  if (_bad_clist != NULL) {
    _paned.add2(_bad_box);
    _bad_box.show();
    _bad_box.pack_start(_bad_label, false, false, 8);
    _bad_label.show();
    request_initial_size(_bad_window, 0, 80);
    _bad_box.pack_start(_bad_window);
    _bad_window.show();
    _bad_window.add(manage(_bad_clist));
    _bad_clist->show();
  }

  // The "Create playlist" check button.
  vbox_.pack_start(_playlist_check);
  _playlist_check.show();

  // The OK button.
  _ok_button.set_usize(80, 36);
  action_area_.pack_start(_ok_button, true, false, 0);
  _ok_button.set_flags(GTK_CAN_DEFAULT);
  _ok_button.grab_default();
  _ok_button.show();
  connect_to_method(_ok_button.clicked, this, &ok_button_clicked);
  
  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();
  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
}



ImportListDialog::Artist::
Artist(const string &name) : _name(name) {
  _artist_id = -1;
}

int ImportListDialog::Artist::
operator < (const Artist &other) const {
  return _name < other._name;
}

ImportListDialog::CD::
CD(const string &title, const Artist *performer, const Artist *writer) :
  _title(title),
  _performer(performer),
  _writer(writer)
{
  _cd_id = -1;
}

int ImportListDialog::CD::
operator < (const CD &other) const {
  if (_title != other._title) {
    return _title < other._title;
  }
  if (_performer != other._performer) {
    return _performer < other._performer;
  }
  return _writer < other._writer;
}

ImportListDialog::Song::
Song() {
  _performer = NULL;
  _writer = NULL;
  _type = -1;
  _cd = NULL;
  _track_num = 0;
}


const ImportListDialog::Artist *
ImportListDialog::define_artist(const string &name) {
  string trim = trim_blanks(name);
  if (trim.empty()) {
    return NULL;
  } else {
    Artists::const_iterator ai = _artists.insert(Artist(name)).first;
    return &(*ai);
  }
}

const ImportListDialog::CD *
ImportListDialog::define_cd(const string &title, const string &performer,
			    const string &writer) {
  CD cd(trim_blanks(title), define_artist(performer), define_artist(writer));
  CDs::const_iterator ci = _cds.insert(cd).first;
  return &(*ci);
}


void ImportListDialog::
collect_songs() {
  int num_rows = _good_files->get_num_rows();
  int num_columns = _good_files->get_num_columns();

  for (int r = 0; r < num_rows; r++) {
    const ImportList::Row &row = _good_files->_rows[r];
    Song *song = new Song;
    song->_relpath = row._relpath;
    song->_type = row._type;

    string cd_title;
    string cd_performer;
    string cd_writer;
    bool has_cd = false;

    for (int c = 0; c < num_columns; c++) {
      const ImportList::Column &column = _good_files->_columns[c];
      const string &data = row._data[c];

      switch (column._type) {
      case ImportList::CT_none:
	break;

      case ImportList::CT_song_name:
	song->_name = trim_blanks(data);
	break;

      case ImportList::CT_performer:
	song->_performer = define_artist(data);
	break;

      case ImportList::CT_writer:
	song->_writer = define_artist(data);
	break;

      case ImportList::CT_track_number:
	song->_track_num = atoi(data.c_str());
	break;

      case ImportList::CT_cd_title:
	cd_title = data;
	has_cd = true;
	break;

      case ImportList::CT_cd_performer:
	cd_performer = data;
	has_cd = true;
	break;

      case ImportList::CT_cd_writer:
	cd_writer = data;
	has_cd = true;
	break;

      default:
	cerr << "Unexpected ImportList::ColumnType " 
	     << column._type << "\n";
	abort();
      }
    }

    if (has_cd) {
      song->_cd = define_cd(cd_title, cd_performer, cd_writer);
    }
    _songs.push_back(song);
  }
}


void ImportListDialog::
import_songs() {
  // Walk through the structures created by collect_songs(), and
  // import them into the database.

  time_t now = time(NULL);

  // First, add all the artists.
  Artists::iterator ai;
  for (ai = _artists.begin(); ai != _artists.end(); ++ai) {
    Artist &import_artist = (Artist &)(*ai);
    if (!import_artist._name.empty()) {
      TableDataProxy db_artist = find_artist(import_artist._name);
      import_artist._artist_id = db_artist.record_db();
    }
  }

  // Now all the CD's.
  CDs::iterator ci;
  for (ci = _cds.begin(); ci != _cds.end(); ++ci) {
    CD &import_cd = (CD &)(*ci);
    TableDataProxy db_cd(TI_cds);
    db_cd[CFI_title] = DataCell(import_cd._title);
    if (import_cd._performer != NULL) {
      db_cd[CFI_performer] = DataCell(import_cd._performer->_artist_id,
				      import_cd._performer->_artist_id >= 0);
    }
    if (import_cd._writer != NULL) {
      db_cd[CFI_writer] = DataCell(import_cd._writer->_artist_id,
				   import_cd._writer->_artist_id >= 0);
    }
    import_cd._cd_id = db_cd.record_db();
  }

  // And now get all the songs.
  Songs::iterator si;
  for (si = _songs.begin(); si != _songs.end(); ++si) {
    Song &import_song = *(*si);
    import_song._db_song = TableDataProxy(TI_songs);
    TableDataProxy &db_song = import_song._db_song;

    if (import_song._performer != NULL) {
      db_song[SFI_performer] = DataCell(import_song._performer->_artist_id,
					import_song._performer->_artist_id >= 0);
    } else if (import_song._cd != NULL && 
	       import_song._cd->_performer != NULL) {
      const Artist *cdperformer = import_song._cd->_performer;
      db_song[SFI_performer] = DataCell(cdperformer->_artist_id,
					cdperformer->_artist_id >= 0);
    }
    if (import_song._writer != NULL) {
      db_song[SFI_writer] = DataCell(import_song._writer->_artist_id,
				     import_song._writer->_artist_id >= 0);
    } else if (import_song._cd != NULL && 
	       import_song._cd->_writer != NULL) {
      const Artist *cdwriter = import_song._cd->_writer;
      db_song[SFI_writer] = DataCell(cdwriter->_artist_id,
				     cdwriter->_artist_id >= 0);
    }
    db_song[SFI_name] = DataCell(import_song._name, !import_song._name.empty());
    if (_disp == D_leave) {
      // If we'll be leaving the files where they are, we can store
      // the file information right now.
      db_song[SFI_file_type] = 
	DataCell(import_song._type, import_song._type >= 0);
      db_song[SFI_volume_id] = DataCell(_volume.get_volume_id());
      db_song[SFI_filename] = DataCell(_volume_prefix + import_song._relpath);
      db_song[SFI_loaded_time] = DataCellTimestamp(now);

    } else {
      // We'll be moving this later, but for now put in a filename so
      // we'll have something to look at.
      db_song[SFI_filename] = DataCell(_volume_prefix + import_song._relpath);
    }

    int song_id = db_song.record_db();

    // Is it a track on a CD?
    if (import_song._cd != NULL) {
      music_db->sql_command
	("INSERT INTO cd_tracks (cd_id, song_id, track_num) VALUES (" +
	 DataCell(import_song._cd->_cd_id) + ", " +
	 DataCell(song_id) + ", " + 
	 DataCell(import_song._track_num) + ")");
    }
  }
}

void ImportListDialog::
build_playlist() {
  Playlist *playlist = new Playlist(TI_songs);
  int num_columns = _good_files->get_num_columns();

  vector<FieldHandle> dbfields;

  // Define all the columns in the playlist.
  for (int c = 0; c < num_columns; c++) {
    const ImportList::Column &column = _good_files->_columns[c];

    if (column._type != ImportList::CT_none) {
      FieldHandle fh;

      switch (column._type) {
      case ImportList::CT_song_name:
	fh = FieldHandle(TI_songs, SFI_name);
	break;

      case ImportList::CT_performer:
	fh = FieldHandle(TI_songs, SFI_performer);
	break;

      case ImportList::CT_writer:
	fh = FieldHandle(TI_songs, SFI_writer);
	break;

      case ImportList::CT_cd_title:
	fh = FieldHandle(TI_cds, CFI_title);
	break;

      case ImportList::CT_cd_performer:
	fh = FieldHandle(TI_cds, CFI_performer);
	break;

      case ImportList::CT_cd_writer:
	fh = FieldHandle(TI_cds, CFI_writer);
	break;
      }

      if (fh._table != TI_no_table) {
	string name = ImportList::get_type_description(column._type);
	DataCell::Type type = ImportList::get_type_datatype(column._type);
	playlist->_columns.push_back(Playlist::Column(name, type));
	dbfields.push_back(fh);
      }      
    }
  } 

  if (dbfields.empty()) {
    playlist->_columns.push_back(Playlist::Column("Filename", 
						  DataCell::DC_string));
    dbfields.push_back(FieldHandle(TI_songs, SFI_filename));
  }

  // Now fill up the rows of the playlist.
  Songs::const_iterator si;
  for (si = _songs.begin(); si != _songs.end(); ++si) {
    Dataset new_data(_data);

    const Song &import_song = *(*si);
    new_data[TI_songs] = import_song._db_song;
    int cd_id = -1;
    if (import_song._cd != NULL) {
      cd_id = import_song._cd->_cd_id;
    }
    new_data[TI_cds] = TableDataProxy(TI_cds, cd_id);

    Playlist::Row row(new_data[TI_songs].get_primary_key());
    if (cd_id >= 0) {
      row._cd_id = cd_id;
    }

    for (int c = 0; c < dbfields.size(); c++) {
      DataCell cell = new_data[dbfields[c]._table][dbfields[c]._field];
      row._column_values.push_back(cell);
    }
    playlist->_rows.push_back(row);
  }

  new PlaylistDialog(_data, -1, playlist);
}


