/* playlist.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "playlist.h"
#include "tableDef.h"
#include "tableField.h"
#include "artist.h"
#include "cd.h"
#include "song.h"
#include "queryResult.h"
#include "dataset.h"
#include "palomaDialog.h"
#include "songDialog.h"
#include "cdDialog.h"
#include "artistDialog.h"
#include "cdReader.h"
#include "string_utils.h"


Playlist::Column::
Column(const string &name, DataCell::Type type, TableID foreign_table) :
  _name(name), _type(type), _foreign_table(foreign_table)
{
  // Look up the column name in our set of known column names in the
  // universe.  If we find a match, assume it to be that column's
  // type.

  const TableField *field = TableDef::find_any_field(name);

  if (field != NULL) {
    _type = field->_field_type;
    _foreign_table = field->_foreign_table;
    if (!field->_label.empty()) {
      _name = field->_label;
    }
  }
}

Playlist::Column::
Column(const Column &copy) :
  _name(copy._name), _type(copy._type), _foreign_table(copy._foreign_table) {
}

Playlist::Column &Playlist::Column::
operator = (const Column &copy) {
  _name = copy._name;
  _type = copy._type;
  _foreign_table = copy._foreign_table;
  return *this;
}

string Playlist::Column::
format_value(const DataCell &value) const {
  if (_type == DataCell::DC_foreign_key) {
    switch (_foreign_table) {
    case TI_artists:
      return get_formatted_artist_name(value.get_foreign(TI_artists));

    case TI_songs:
      return get_formatted_song_name(value.get_foreign(TI_songs));

    case TI_cds:
      return get_formatted_cd_name(value.get_foreign(TI_cds));

    case TI_combo_items:
      return value.get_foreign(TI_combo_items)[CIFI_label].get_string();
    }
  }

  return DataCell(value, _type).get_text();
}

Playlist::Row::
Row(int primary_key, int cd_id) : _primary_key(primary_key), _cd_id(cd_id)
{
}

Playlist::AcceptDialogData::
AcceptDialogData(PalomaDialog *dialog, int row) :
  _dialog(dialog), _row(row) {
}

Playlist::
Playlist(TableID primary_table) : _primary_table(primary_table) {
  _clist = NULL;
  _cd_reader = NULL;
  _owns_dialogs = true;
  _next_form_id = -1;
}

Playlist::
~Playlist() {
  // Delete all of the dialogs we own.
  Dialogs::iterator di;
  for (di = _dialogs.begin(); di != _dialogs.end(); ++di) {
    delete (*di);
  }

#ifdef HAVE_CDAUDIO
  if (_cd_reader != NULL) {
    delete _cd_reader;
  }
#endif // HAVE_CDAUDIO

  if (_clist != NULL) {
    delete _clist;
  }
}

// Once a playlist is constructed, add rows to it from the given
// query.  This query must be compatible with the one that constructed
// the playlist.
void Playlist::
append_data(const QueryResult &qr) {
  int num_rows = qr.get_num_rows();
  int num_columns = qr.get_num_columns();

  if (num_columns == 0) {
    return;
  }

  // Normally, we display all columns in the result, including the
  // first column.
  int start_at = 0;
  int extra_cd_id = -1;

  if (_columns.empty()) {
    // If we don't have any columns yet, this must be the first query
    // to be added to the playlist.  Use it to define the columns.

    // First, figure out what kind of table the query is based on.  We
    // use the first column of the query to determine this; it should
    // be the primary key of any of our three possible tables.  If it
    // is not, then we are a non-table playlist, and while the user
    // can inspect the data as a general SQL query, he/she can't play
    // any songs.
    
    const TableField *field =
      TableDef::find_any_field(qr.get_column_name(0));

    if (field != NULL) {
      if (field->_field_type == DataCell::DC_foreign_key) {
	switch (field->_foreign_table) {
	case TI_artists:
	case TI_cds:
	case TI_songs:
	  _primary_table = field->_foreign_table;
	}
      }
    }

    // Normally, if the first column is used to define the table, we
    // don't include the first column in the result.
    if (_primary_table != TI_no_table) {
      start_at = 1;
    }

    // As a special exception, if the first column was a song ID and
    // the second column is a CD ID, we have a song/CD pair.
    if (_primary_table == TI_songs && num_columns >= 2) {
      const TableField *field =
	TableDef::find_any_field(qr.get_column_name(1));
      if (field != NULL && 
	  field->_field_type == DataCell::DC_foreign_key &&
	  field->_foreign_table == TI_cds) {
	start_at = 2;
	extra_cd_id = 1;
      }
    }

    if (start_at >= num_columns) {
      // However, if we don't have any columns left, we do include
      // these special columns.
      start_at = 0;
    }

    for (int c = start_at; c < num_columns; c++) {
      _columns.push_back(Column(qr.get_column_name(c),
				qr.get_column_type(c)));
    }

  } else {
    // If we already have columns, this query is an addendum to the
    // one(s) before.  Make sure it is compatible.

    start_at = num_columns - _columns.size();
    if (start_at == 2) {
      extra_cd_id = 1;
    }

    assert(num_columns - start_at == _columns.size());
  }

  // Now add in all the data.
  for (int r = 0; r < num_rows; r++) {
    int primary_key = -1;
    if (_primary_table != TI_no_table) {
      primary_key = qr.get_cell(r, 0).get_int();
    }
    Row row(primary_key);
    if (extra_cd_id >= 0) {
      row._cd_id = qr.get_cell(r, extra_cd_id).get_int();
    }
    
    for (int c = start_at; c < num_columns; c++) {
      row._column_values.push_back(qr.get_cell(r, c));
    }
    
    _rows.push_back(row);
  }
}

string Playlist::
format_value(int r, int c) const {
  return _columns[c].format_value(get_row(r)->_column_values[c]);
}


Gtk_CList *Playlist::
make_clist() {
  assert(!_columns.empty());
  int num_columns = _columns.size();
  const char *titles[num_columns];
  for (int i = 0; i < num_columns; i++) {
    titles[i] = _columns[i]._name.c_str();
  }

  // We can't assign the _clist member yet, because we need
  // format_value() below to look up its value in the _rows array.
  // Once we assign a value to _clist it will try to look there
  // instead!
  Gtk_CList *clist = new Gtk_CList(num_columns, titles);

  vector<const char *> row_text(num_columns);
  vector<string> row_strings(num_columns);

  // It seems that trying to stuff more than about 2000 rows into a
  // CList is generally a bad idea.
  int num_rows = min((int)_rows.size(), 2000);

  for (int r = 0; r < num_rows; r++) {
    bool all_empty = true;
    for (int c = 0; c < num_columns; c++) {
      row_strings[c] = trim_blanks(format_value(r, c));
      all_empty = all_empty && row_strings[c].empty();
      row_text[c] = row_strings[c].c_str();
    }
    if (all_empty) {
      row_text[0] = "(Unnamed)";
    }

    clist->append(&row_text[0]);
    clist->set_row_data(r, &_rows[r]);
  }

  connect_to_method(clist->button_press_event, this, &got_clist_clicked);

  _clist = clist;
  return _clist;
}

Gtk_CList *Playlist::
get_clist() const {
  return _clist;
}

void Playlist::
update_clist(int row) {
  assert(_clist != NULL);
  int num_columns = _columns.size();
  for (int c = 0; c < num_columns; c++) {
    _clist->set_text(row, c, format_value(row, c));
  }
}

int Playlist::
get_num_rows() const {
  if (_clist == NULL) {
    return _rows.size();

  } else {
    // If we have a CList, that defines the length of the playlist.
    // (It might be smaller than the size of the _rows vector, because
    // of arbitrary limits imposed on the maximum length of a CList.)
    return _clist->rows();
  }
}

const Playlist::Row *Playlist::
get_row(int row) const {
  if (_clist == NULL) {
    assert(row >= 0 && row < _rows.size());
    return &_rows[row];
  } else {
    assert(row >= 0 && row < _clist->rows());
    return (Row *)_clist->get_row_data(row);
  }
}

void Playlist::
popup_dialog(int row) {
  assert(row >= 0 && row < get_num_rows());

  Dataset data;
  fill_dialog_data(row, data);

  PalomaDialog *dialog = NULL;

  switch (_primary_table) {
  case TI_songs:
    dialog = new SongDialog(data, _next_form_id, _cd_reader, row);
    break;

  case TI_cds:
    dialog = CDDialog::make_dialog(data, _next_form_id);
    break;

  case TI_artists:
    dialog = new ArtistDialog(data, _next_form_id);
    break;
  }

  if (dialog != NULL) {
    connect_to_method(dialog->accepted, this, &got_dialog_accept, 
		      AcceptDialogData(dialog, row));

    dialog_created(row, dialog);

    if (_owns_dialogs) {
      connect_to_method(dialog->deleted, this, &got_dialog_delete, 
			(PalomaDialog *)dialog);
      _dialogs.insert(dialog);
    }
  }
}

TableID Playlist::
get_primary_table() const {
  return _primary_table;
}

void Playlist::
set_owns_dialogs(bool flag) {
  _owns_dialogs = flag;
}

void Playlist::
set_next_form_id(int next_form_id) {
  _next_form_id = next_form_id;
}

void Playlist::
set_cd_reader(CDReader *cd_reader) {
  _cd_reader = cd_reader;
}

CDReader *Playlist::
get_cd_reader() const {
  return _cd_reader;
}

TableDataProxy Playlist::
get_table_data(int row) {
  if (_primary_table == TI_no_table) {
    return TableDataProxy();
  } else {
    return TableDataProxy(_primary_table, get_row(row)->_primary_key);
  }
}

// The default behavior for fill_dialog_data() is to look the data up
// in the database.
void Playlist::
fill_dialog_data(int row, Dataset &data) {
  if (_primary_table == TI_no_table) {
    return;
  }
  const TableDataProxy &table_data = get_table_data(row);
  data[_primary_table] = table_data;
  int primary_key = table_data.get_primary_key();

  switch (_primary_table) {
  case TI_songs:
    data[TI_songs_extension] =
      TableDataProxy(TI_songs_extension, primary_key);
    break;

  case TI_artists:
    data[TI_artists_extension] =
      TableDataProxy(TI_artists_extension, primary_key);
    break;

  case TI_cds:
    data[TI_cds_extension] =
      TableDataProxy(TI_cds_extension, primary_key);
    break;
  }
}

// The default behavior for dialog_created() is to do nothing.
void Playlist::
dialog_created(int, PalomaDialog *) {
}

// The default behavior for dialog_accepted() is to record the dialog
// into the database.
void Playlist::
dialog_accepted(int, PalomaDialog *dialog) {
  dialog->record_db();
}


gint Playlist::
got_clist_clicked(GdkEventButton *event) {
  // If we get a double-click button event, we'll pop up a dialog for
  // the indicated row.
  if (event->button == 1 && event->type == GDK_2BUTTON_PRESS) {
    int row, column;
    _clist->get_selection_info(event->x, event->y, &row, &column);
    if (row >= 0 && row < get_num_rows()) {
      popup_dialog(row);
    }
  }
  return false;
}

void Playlist::
got_dialog_accept(AcceptDialogData add) {
  dialog_accepted(add._row, add._dialog);
}

void Playlist::
got_dialog_delete(PalomaDialog *dialog) {
  _dialogs.erase(dialog);
}

